
import re
from .exceptions import ParsionException


class ParsionLexerError(ParsionException):
    def __init__(self, message, input, pos):
        super().__init__(message)
        self.input = input
        self.pos = pos

    def __str__(self):
        """
        Format lexer error

        >>> str(ParsionLexerError("msg", "my input", 3))
        "msg (pos 3 in 'my input')"
        """
        return f'{self.args[0]} (pos {self.pos} in {self.input!r})'


class ParsionToken:
    def __init__(self, name, value, start, end):
        self.name = name
        self.value = value
        self.start = start
        self.end = end

    def __str__(self):
        """
        Format token as string

        >>> str(ParsionToken("my_name", "my_value", 12, 14))
        "[@ 12 my_name: 'my_value']"

        >>> str(ParsionToken("my_name", None, 12, 14))
        '[@ 12 my_name]'
        """
        if self.value is None:
            return f'[@{self.start:>3} {self.name}]'
        else:
            return f'[@{self.start:>3} {self.name}: {self.value!r}]'

    def ignore(self):
        return self.name is None


class ParsionEndToken(ParsionToken):
    def __init__(self, pos):
        super().__init__('$END', '$END', pos, pos)


class ParsionLexer:
    def __init__(self, rules):
        self.rules = [
            (name, re.compile(regexp), handler)
            for (name, regexp, handler)
            in rules
        ]

    def next_token(self, input, pos):
        for name, regexp, handler in self.rules:
            m = regexp.match(input, pos)
            if m is not None:
                return ParsionToken(
                    name,
                    handler(m.group(1)),
                    m.start(1),
                    m.end(1)
                )
        return None

    def tokenize(self, input):
        pos = 0
        while pos < len(input):
            token = self.next_token(input, pos)
            if token is None:
                raise ParsionLexerError('Invalid input', input, pos)
            pos = token.end
            if not token.ignore():
                yield token
        yield ParsionEndToken(pos)

    def get_token_set(self):
        return {
            rule[0]
            for rule
            in self.rules
            if rule[0] is not None
        }.union({'$END'})



from .exceptions import ParsionGeneratorError


def _noset(obj):
    """
    Remove sets from obj, only for doctests

    It looses the use case of doctests, but when an object returns a set, the
    order is random and can't be tested otherwise

    >>> _noset(12)
    12

    >>> _noset({1, 4, 3})
    [1, 3, 4]

    >>> _noset((4, 3, {4, 2}))
    (4, 3, [2, 4])

    >>> _noset([1, 4, {1, 3}, 3])
    [1, 4, [1, 3], 3]
    """
    if isinstance(obj, set):
        return sorted(_noset(x) for x in obj)
    elif isinstance(obj, tuple):
        return tuple(_noset(x) for x in obj)
    elif isinstance(obj, list):
        return [_noset(x) for x in obj]
    else:
        return obj


class ParsionFSMMergeError(Exception):
    pass


class ParsionFSMGrammarRule:
    def __init__(self, id, name, gen, rulestr):
        self.id = id
        self.name = name
        self.gen = gen

        parts = rulestr.split(' ')
        self.attrtokens = [part[0] != '_' for part in parts]
        self.parts = [part[1:] if part[0] == '_' else part for part in parts]

        # This class will never change value. Precalculate hash
        self.hash = hash((
            type(self).__name__,
            self.gen,
            self.name,
            sum(hash(t) for t in self.attrtokens)
        ))

    def get(self, idx, default=None):
        if idx < len(self.parts):
            return self.parts[idx]
        else:
            return default

    def export(self):
        return (self.gen, self.name, self.attrtokens)

    def _tupleize(self):
        """
        Get a tuple of all relevant parameters, for usage in __eq__ and __lt__

        >>> ParsionFSMGrammarRule(12, 'name', 'gen', 'lhs _op rhs')._tupleize()
        ('name', 'gen', ['lhs', 'op', 'rhs'], [True, False, True])
        """
        return (self.name or '', self.gen, self.parts, self.attrtokens)

    def __hash__(self):
        return self.hash

    def __lt__(self, other):
        return self._tupleize() < other._tupleize()

    def __eq__(self, other):
        return self._tupleize() == other._tupleize()

    def __str__(self):  # pragma: no cover
        name = f'{self.name}:' if self.name is not None else ''
        return f'{name:<12} {self.gen:<10} = {" ".join(self.parts)}'


class ParsionFSMItem:
    def __init__(self, rule, follow, pos=0):
        self.rule = rule
        self.pos = pos
        self.follow = set(follow)

        # This class will never change value. Precalculate hash
        self.hash = hash((
            type(self).__name__,
            self.rule,
            self.pos,
            sum(hash(x) for x in self.follow)
        ))

    def __str__(self):  # pragma: no cover
        name = f'{self.rule.name}:' if self.rule.name is not None else ''
        fmt_parts = [
            part if i != self.pos else f'>{part}<'
            for i, part in enumerate(self.rule.parts)
        ]
        return f'{name:<12} {self.rule.gen:<10} = {" ".join(fmt_parts)}'

    def _tupleize(self):
        """
        Get a tuple of all relevant parameters, for usage in __eq__ and __lt__
        """
        return (self.rule, self.pos, self.follow)

    def __hash__(self):
        return self.hash

    def __lt__(self, other):
        return self._tupleize() < other._tupleize()

    def __eq__(self, other):
        return self._tupleize() == other._tupleize()

    def get_next(self):
        """
        Get next two symbols from an item

        >>> rule = ParsionFSMGrammarRule(12, 'name', 'gen', 'lhs _op rhs')

        >>> ParsionFSMItem(rule, {'fa', 'fb'}, 0).get_next()
        ('lhs', {'op'})

        >>> ParsionFSMItem(rule, {'fa', 'fb'}, 1).get_next()
        ('op', {'rhs'})

        >>> _noset(ParsionFSMItem(rule, {'fa', 'fb'}, 2).get_next())
        ('rhs', ['fa', 'fb'])

        >>> ParsionFSMItem(rule, {'fa', 'fb'}, 3).get_next() is None
        True
        """
        n = self.rule.get(self.pos)
        if n is None:
            return None
        f = self.rule.get(self.pos + 1)
        if f is None:
            f = self.follow
        else:
            f = {f}
        return n, f

    def is_complete(self):
        return self.rule.get(self.pos) is None

    def take(self, sym):
        if self.rule.get(self.pos) == sym:
            return ParsionFSMItem(self.rule, self.follow, self.pos + 1)
        else:
            return None

    def is_mergable(self, other):
        return self.rule == other.rule and self.pos == other.pos

    def merge(self, other):
        """
        Merge two identical items but with different follows, and return the
        combined

        If the two items are not compatible, throw an error

        >>> rule = ParsionFSMGrammarRule(12, 'name', 'gen', 'lhs _op rhs')
        >>> a = ParsionFSMItem(rule, {'fa', 'fb'}, 0)
        >>> b = ParsionFSMItem(rule, {'fb', 'fc'}, 0)
        >>> sorted(a.merge(b).follow)
        ['fa', 'fb', 'fc']

        >>> a0 = ParsionFSMItem(rule, {'fa', 'fb'}, 0)
        >>> b1 = ParsionFSMItem(rule, {'fb', 'fc'}, 1)
        >>> a0.merge(b1)
        Traceback (most recent call last):
        ...
        parsion.parsegen.ParsionFSMMergeError

        """
        if not self.is_mergable(other):
            raise ParsionFSMMergeError()
        return ParsionFSMItem(
            self.rule,
            self.follow.union(other.follow),
            self.pos
        )


class ParsionFSMState:

    def __init__(self, items):
        self.items = set(items)
        self.hash = sum(hash(it) for it in self.items)

    def next_syms(self):
        return set(
            it.get_next()[0]
            for it
            in self.items
            if not it.is_complete()
        )

    def reductions(self):
        return [it for it in self.items if it.is_complete()]

    def take(self, sym):
        result = []
        for item in self.items:
            next_item = item.take(sym)
            if next_item is not None:
                result.append(next_item)
        return result

    def __hash__(self):
        return self.hash

    def __str__(self):  # pragma: no cover
        return "\n".join(str(it) for it in self.items)

    def __eq__(self, other):
        return self.items == other.items


class ParsionFSM:
    def __init__(self, grammar_rules):
        self.error_rules = {
            gen: name
            for (name, gen, rulestr)
            in grammar_rules
            if rulestr == '$ERROR'
        }

        no_error_rules = [
            (name, gen, rulestr)
            for (name, gen, rulestr)
            in grammar_rules
            if rulestr != '$ERROR'
        ]

        self.grammar = [
            ParsionFSMGrammarRule(
                0,
                None,
                '$ENTRY',
                'entry _$END'
            )
        ] + [
            ParsionFSMGrammarRule(id + 1, name, gen, rulestr)
            for id, (name, gen, rulestr)
            in enumerate(no_error_rules)
        ]

        self._build_sym_set()
        self._calculate_firsts()
        self._build_states()

    def _get_rules_by_gen(self, gen):
        return [rule for rule in self.grammar if rule.gen == gen]

    def _add_state(self, state):
        state_id = self.state_ids.get(state)
        if state_id is None:
            state_id = len(self.states)
            self.state_ids[state] = state_id
            self.states.append(state)
            self.table.append({})
        return state_id

    def _build_sym_set(self):
        self.sym_set = set()
        for rule in self.grammar:
            self.sym_set.add(rule.gen)
            self.sym_set.update(rule.parts)

    def _calculate_firsts(self):
        rule_firsts = {rule.gen: rule.parts[0] for rule in self.grammar}
        self.firsts = {}
        for sym in self.sym_set:
            first_set = set()
            cur_sym = sym
            while cur_sym not in first_set:
                first_set.add(cur_sym)
                if cur_sym in rule_firsts:
                    cur_sym = rule_firsts[cur_sym]
            self.firsts[sym] = first_set

    def _get_first(self, syms):
        result = set()
        for sym in syms:
            result.update(self.firsts.get(sym, set()))
        return result

    def _get_closure(self, items):
        """
        Get a closure from list of items

        A closure is the input items, but also populated with new items from
        grammars, which generates the next symbol of the incoming list of items
        """

        all_items = {}
        queue = []

        for item in items:
            queue.append(item)

        # Resolve all sub items
        while len(queue) > 0:
            it = queue.pop()

            key = (it.rule, it.pos)
            if key in all_items:
                old_it = all_items[key]
                new_it = old_it.merge(it)
                if new_it == old_it:
                    continue
                all_items[key] = new_it
            else:
                all_items[key] = it

            if not it.is_complete():
                sym, follow = it.get_next()
                follow_first = self._get_first(follow)
                for rule in self._get_rules_by_gen(sym):
                    queue.append(ParsionFSMItem(rule, follow_first))

        return sorted(all_items.values())

    def _build_states(self):
        self.states = []
        self.table = []
        self.state_ids = {}
        self.error_handlers = {}

        self._add_state(
            ParsionFSMState(self._get_closure([
                ParsionFSMItem(
                    self.grammar[0],
                    set()
                )
            ]))
        )

        state_queue = [0]
        processed = set()

        while len(state_queue) > 0:
            state_id = state_queue.pop(0)
            if state_id in processed:
                continue
            state = self.states[state_id]
            processed.add(state_id)

            # Check if state can have an error handler
            error_handlers = {}
            for it in state.items:
                if it.rule.gen in self.error_rules:
                    for sym in it.follow:
                        if sym in error_handlers:
                            raise ParsionGeneratorError(
                                f'{it.rule.gen}: {sym} handler already defined'
                            )
                        error_handlers[sym] = (
                            it.rule.gen, self.error_rules[it.rule.gen])
            if error_handlers != {}:
                self.error_handlers[state_id] = error_handlers

            # Process rules
            for sym in state.next_syms():
                next_id = self._add_state(ParsionFSMState(
                    self._get_closure(state.take(sym))))
                state_queue.append(next_id)
                self.table[state_id][sym] = ('s', next_id)

            for it in state.reductions():
                for sym in it.follow:
                    if sym in self.table[state_id]:
                        raise ParsionGeneratorError("Shift/Reduce conflict")
                    self.table[state_id][sym] = ('r', it.rule.id)

    def export(self):
        return (
            [g.export() for g in self.grammar],
            self.table,
            self.error_handlers
        )



from .exceptions import ParsionParseError, ParsionInternalError


class ParsionParser:
    def __init__(self, parse_grammar, parse_table, error_handlers):
        self.parse_grammar = parse_grammar
        self.parse_table = parse_table
        self.error_handlers = error_handlers

    def _call_reduce(self, obj, goal, accepts, parts):
        args = [p[0] for a, p in zip(accepts, parts) if a]

        if goal is None:
            assert len(args) == 1
            return args[0]
        else:
            return getattr(obj, goal)(*args)

    def _call_error_handler(self, obj, handler, error_stack, error_tokens):
        return getattr(obj, handler)(error_stack, error_tokens)

    def parse(self, input, handlerobj):
        tokens = [(tok.name, tok.value) for tok in input]
        stack = [('START', 0)]

        while len(tokens) > 0:
            tok_name, tok_value = tokens[0]
            cur_state = stack[-1][1]
            if tok_name not in self.parse_table[cur_state]:
                # Unexpected token, do error recovery
                try:
                    # First, pop stack until error handler
                    error_stack = []
                    while stack[-1][1] not in self.error_handlers:
                        error_stack.append(stack.pop())

                    error_handlers = self.error_handlers[stack[-1][1]]

                    error_tokens = []
                    while tokens[0][0] not in error_handlers:
                        error_tokens.append(tokens.pop(0))

                    # Call error handler, mimic a reduce operation
                    error_gen, error_handler = error_handlers[tokens[0][0]]
                    value = self._call_error_handler(
                        handlerobj,
                        error_handler,
                        error_stack,
                        error_tokens
                    )
                    tokens.insert(0, (error_gen, value))
                except IndexError:
                    expect_toks = ",".join(self.parse_table[cur_state].keys())
                    raise ParsionParseError(
                        f'Unexpected {tok_name}, expected {expect_toks}')
            else:
                op, id = self.parse_table[cur_state][tok_name]
                if op == 's':
                    # shift
                    tokens.pop(0)
                    stack.append((tok_value, id))
                elif op == 'r':
                    # reduce
                    gen, goal, accepts = self.parse_grammar[id]
                    tokens.insert(0, (
                        gen,
                        self._call_reduce(
                            handlerobj,
                            goal,
                            accepts,
                            stack[-len(accepts):]
                        )
                    ))
                    stack = stack[:-len(accepts)]
                else:
                    raise ParsionInternalError(
                        'Internal error: neigher shift nor reduce')

        # Stack contains three elements:
        #  0. ('START', ...) - bootstrap
        #  1. ('entry', ...) - excpeted result
        #  2. ('END', ...)   - terminination
        # Therefore, pick out entry value and return
        return stack[1][0]

    def print(self):  # pragma: no cover
        from tabulate import tabulate

        def _print_header(header):
            print("")
            print(f"{header}")

        _print_header("Lexer")
        print(tabulate(
            [
                (name, regexp.pattern)
                for name, regexp, handler in self.lex.rules
            ],
            tablefmt='simple_outline'
        ))

        _print_header("FSM grammar")
        print(tabulate(
            self.parse_grammar,
            headers=('generate', 'goal', 'parts'),
            tablefmt='simple_outline',
            showindex='always'
        ))
        _print_header("FSM states")

        print(tabulate(
            [
                {
                    k: " ".join(str(p) if p is not None else '-' for p in v)
                    for k, v in st.items()
                }
                for st in self.parse_table
            ],
            headers='keys',
            tablefmt='simple_outline',
            showindex='always'
        ))



from .core import Parsion, ParsionStatic
from .lex import ParsionLexer, ParsionEndToken, ParsionLexerError, ParsionToken
from .parser import ParsionParser
from .exceptions import ParsionException, ParsionGeneratorError, \
    ParsionInternalError, ParsionSelfCheckError, ParsionParseError

__all__ = [
    'Parsion',
    'ParsionStatic',
    'ParsionLexer',
    'ParsionParseError',
    'ParsionLexerError',
    'ParsionToken',
    'ParsionEndToken',
    'ParsionParser',
    'ParsionParseError',
    'ParsionException',
    'ParsionGeneratorError',
    'ParsionInternalError',
    'ParsionSelfCheckError'
]



class ParsionException(Exception):
    pass


class ParsionGeneratorError(ParsionException):
    pass


class ParsionInternalError(ParsionException):
    pass


class ParsionSelfCheckError(ParsionException):
    pass


class ParsionParseError(Exception):
    pass



from .lex import ParsionLexer
from .parser import ParsionParser
from .parsegen import ParsionFSM


class ParsionBase:
    LEXER_RULES = []
    SELF_CHECK = True

    def __init__(self, lexer, parser):
        self.lexer = lexer
        self.parser = parser
        if self.SELF_CHECK:
            self._self_check()

    def parse(self, input):
        tokens = self.lexer.tokenize(input)
        return self.parser.parse(tokens, self)

    def _self_check(self):
        from .self_check import run_self_check
        run_self_check(self)

    def entry(self, v):
        return v


class Parsion(ParsionBase):
    GRAMMAR_RULES = []

    def __init__(self):
        (
            self.parse_grammar,
            self.parse_table,
            self.error_handlers
        ) = ParsionFSM(self.GRAMMAR_RULES).export()

        super().__init__(
            ParsionLexer(self.LEXER_RULES),
            ParsionParser(
                self.parse_grammar,
                self.parse_table,
                self.error_handlers
            )
        )


class ParsionStatic(ParsionBase):
    STATIC_GRAMMAR = None
    STATIC_TABLE = None
    STATIC_ERROR_HANDLERS = None

    def __init__(self):
        super().__init__(
            ParsionLexer(self.LEXER_RULES),
            ParsionParser(
                self.STATIC_GRAMMAR,
                self.STATIC_TABLE,
                self.STATIC_ERROR_HANDLERS
            )
        )



from .exceptions import ParsionSelfCheckError


def _self_check_handlers(par):
    """
    Startup check for handlers

    Check that all handlers are implemented with correct amount of arguments
    """
    # Keep inspect in this functions, so it can easily be disabled
    import inspect

    expected_funcs = {}

    # Check all reduce handlers are accessable
    for i, (gen, goal, accepts) in enumerate(par.parser.parse_grammar):
        argc = sum(1 for a in accepts if a)
        if goal is None:
            if argc != 1:
                raise ParsionSelfCheckError(
                    f'No handler for rule #{i} (gen: {gen}), but {argc} args'
                )
        else:
            expected_funcs[goal] = argc

    # Check all error handlers are implemented
    for error_handlers in par.parser.error_handlers.values():
        for gen, handler in error_handlers.values():
            expected_funcs[handler] = 2  # error_stack, error_tokens

    # Check all reduce handlers are accessable
    for goal, arg_count in expected_funcs.items():
        try:
            handler = inspect.signature(getattr(par, goal))

            # Count minimum and maximum number of arguments
            param_count_min = 0
            param_count_max = 0
            for p in handler.parameters.values():
                if p.kind in {p.POSITIONAL_ONLY, p.POSITIONAL_OR_KEYWORD}:
                    # Positional attribute, with or without default value
                    param_count_max += 1
                    if p.default == p.empty:
                        param_count_min += 1
                if p.kind in {p.VAR_POSITIONAL}:
                    # Variable attribute: *args
                    param_count_max = None
                    # No more paramters can come afterwards
                    break

            if arg_count < param_count_min or (
                param_count_max is not None and param_count_max < arg_count
            ):
                if param_count_min == param_count_max:
                    count_str = f'{param_count_min}'
                elif param_count_max is None:
                    count_str = f'>={param_count_min}'
                else:
                    count_str = \
                        f'between {param_count_min} and {param_count_max}'
                raise ParsionSelfCheckError(
                    f"{goal}: expected {arg_count} args, has {count_str}")
        except AttributeError:
            raise ParsionSelfCheckError(f"No handler defined for {goal}")


def run_self_check(par):
    _self_check_handlers(par)
